@@ -3,8 +3,8 @@
 import gymnasium as gym
 from shutil import rmtree
 
-evolutionary = True
-continuous_actions = True
+evolutionary = False
+continuous_actions = False
 
 env = gym.make(
     'LunarLander-v3',
