@@ -6,7 +6,7 @@ from math import ceil
 from pathlib import Path
 from copy import deepcopy
 from functools import partial, wraps
-from collections import deque, namedtuple
+from collections import namedtuple
 from random import random
 
 import numpy as np
@@ -16,7 +16,7 @@ import torch
 from torch import nn, tensor, Tensor, is_tensor, cat, stack, zeros, ones, full, arange
 import torch.distributed as dist
 import torch.nn.functional as F
-from torch.nn import Module
+from torch.nn import Module, GRU
 from torch.utils.data import TensorDataset, DataLoader
 from torch.distributions import Categorical, Normal
 from torch.utils._pytree import tree_map
@@ -723,6 +723,7 @@ class WorldModelActorCritic(Module):
         actions = None,
         rewards = None,
         reward_dropout: bool | Tensor | None = None,
+        prev_action_dropout: bool | Tensor | None = None,
         next_actions = None,
         latent_gene = None,
         world_model_embed_mult = 1.,
@@ -736,6 +737,16 @@ class WorldModelActorCritic(Module):
 
         if exists(actions):
             action_embeds = self.action_embeds(actions)
+
+            if not is_tensor(prev_action_dropout):
+                prev_action_dropout = tensor(prev_action_dropout, device = device)
+
+            if prev_action_dropout.ndim == 0:
+                prev_action_dropout = repeat(prev_action_dropout, '-> b', b = batch)
+
+            if exists(prev_action_dropout):
+                action_embeds = multiply('b ..., b -> b ...', action_embeds, (~prev_action_dropout).float())
+
             sum_embeds = sum_embeds + action_embeds
 
         if exists(rewards):
@@ -773,6 +784,7 @@ class WorldModelActorCritic(Module):
             *args,
             **kwargs,
             cache = cache,
+            input_not_include_cache = True,
             sum_embeds = sum_embeds,
             prepend_embeds = prepend_embeds,
             return_embeddings = True,
@@ -876,6 +888,11 @@ class RSNorm(Module):
 
         return normalized * std + mean
 
+    @torch.no_grad()
+    def forward_eval(self, x):
+        self.eval()
+        return self.forward(x)
+
     def forward(
         self,
         x
@@ -953,7 +970,7 @@ class Agent(Module):
         betas,
         lam,
         gamma,
-        beta_s,
+        entropy_weight,
         regen_reg_rate,
         cautious_factor,
         eps_clip,
@@ -973,9 +990,10 @@ class Agent(Module):
             num_selected = 2,
             tournament_size = 2
         ),
+        world_model_attn_dim_head = 16,
+        world_model_heads = 4,
+        world_model_attn_hybrid_gru = False,
         world_model: dict = dict(
-            attn_dim_head = 16,
-            heads = 4,
             depth = 4,
             attn_gate_values = True,
             add_value_residual = True,
@@ -1011,6 +1029,8 @@ class Agent(Module):
         if evolutionary:
             self.gene_pool = LatentGenePool(**latent_gene_pool)
 
+        maybe_gru = GRU(hidden_dim, world_model_attn_dim_head * world_model_heads, batch_first = True) if world_model_attn_hybrid_gru else None
+
         self.model = WorldModelActorCritic(
             num_actions = num_actions,
             num_set_actions = num_set_actions,
@@ -1019,7 +1039,7 @@ class Agent(Module):
             critic_dim_pred = critic_pred_num_bins,
             critic_min_max_value = reward_range,
             state_dim = state_dim,
-            entropy_weight = beta_s,
+            entropy_weight = entropy_weight,
             eps_clip = eps_clip,
             value_clip = value_clip,
             evolutionary = evolutionary,
@@ -1035,6 +1055,10 @@ class Agent(Module):
                     attn_dropout = dropout,
                     ff_dropout = dropout,
                     verbose = False,
+                    attn_hybrid_fold_axial_dim = 1,
+                    attn_hybrid_module = maybe_gru,
+                    attn_dim_head = world_model_attn_dim_head,
+                    heads = world_model_heads,
                     **world_model
                 )
             ),
@@ -1074,7 +1098,7 @@ class Agent(Module):
 
         self.lam = lam
         self.gamma = gamma
-        self.beta_s = beta_s
+        self.entropy_weight = entropy_weight
 
         self.eps_clip = eps_clip
         self.value_clip = value_clip
@@ -1135,6 +1159,7 @@ class Agent(Module):
         episode_lens,
         gene_ids,
         reward_dropouts,
+        prev_action_dropouts,
         fitnesses = None
     ):
 
@@ -1188,7 +1213,8 @@ class Agent(Module):
             is_boundaries,
             gene_ids,
             episode_lens,
-            reward_dropouts
+            reward_dropouts,
+            prev_action_dropouts
         )
 
         if is_distributed():
@@ -1224,7 +1250,8 @@ class Agent(Module):
                 dones,
                 gene_ids,
                 episode_lens,
-                reward_dropout
+                reward_dropout,
+                prev_action_dropout
              ) in dl:
 
                 latent_gene = None
@@ -1246,9 +1273,7 @@ class Agent(Module):
 
                 states_with_rewards, inverse_pack = pack_with_inverse((states, rewards), 'b n *')
 
-                with torch.no_grad():
-                    self.rsnorm.eval()
-                    states_with_rewards = self.rsnorm(states_with_rewards)
+                states_with_rewards = self.rsnorm.forward_eval(states_with_rewards)
 
                 states, rewards = inverse_pack(states_with_rewards)
 
@@ -1256,6 +1281,7 @@ class Agent(Module):
                     states,
                     rewards = rewards,
                     reward_dropout = reward_dropout,
+                    prev_action_dropout = prev_action_dropout,
                     actions = prev_actions,
                     latent_gene = latent_gene,
                     next_actions = actions, # prediction of the next state needs to be conditioned on the agent's chosen action on that state, and will make the world model interactable
@@ -1379,9 +1405,7 @@ class Agent(Module):
         if not has_reward:
             state = F.pad(state, (0, 1), value = 0.)
 
-        with torch.no_grad():
-            self.rsnorm.eval()
-            normed_state_with_reward = self.rsnorm(state)
+        normed_state_with_reward = self.rsnorm.forward_eval(state)
 
         normed_state = normed_state_with_reward[..., :-1]
 
@@ -1418,6 +1442,7 @@ class Learner(Module):
         evolve_every = 10,
         evolve_after_step = 20,
         reward_dropout_prob = 0.5,
+        action_dropout_prob = 0.5,
         curiosity_reward_weight = 0.,
         latent_gene_pool: dict | None = None,
         max_timesteps = 500,
@@ -1429,7 +1454,7 @@ class Learner(Module):
         gamma = 0.99,
         eps_clip = 0.2,
         value_clip = 0.4,
-        beta_s = .01,
+        entropy_weight = .01,
         regen_reg_rate = 1e-4,
         cautious_factor = 0.1,
         epochs = 3,
@@ -1480,7 +1505,7 @@ class Learner(Module):
             betas = betas,
             lam = lam,
             gamma = gamma,
-            beta_s = beta_s,
+            entropy_weight = entropy_weight,
             regen_reg_rate = regen_reg_rate,
             cautious_factor = cautious_factor,
             eps_clip = eps_clip,
@@ -1529,6 +1554,10 @@ class Learner(Module):
 
         self.reward_dropout_prob = reward_dropout_prob
 
+        # action dropout for world models
+
+        self.action_dropout_prob = action_dropout_prob
+
         # weight for curiosity reward, based on entropy of state prediction
 
         self.curiosity_reward_weight = curiosity_reward_weight
@@ -1557,10 +1586,11 @@ class Learner(Module):
 
         agent, device, num_episodes_per_update, is_main = self.agent, self.device, self.num_episodes_per_update, self.accelerator.is_main_process
 
-        memories = deque([])
+        memories = []
         episode_lens = []
         gene_ids = []
         one_update_reward_dropouts = []
+        one_update_action_dropouts = []
 
         if exists(seed):
             torch.manual_seed(seed)
@@ -1594,6 +1624,12 @@ class Learner(Module):
 
             reward_dropouts = torch.rand((num_episodes_per_update,), device = device) < self.reward_dropout_prob
 
+            # precompute prev action dropouts
+
+            action_dropouts = torch.rand((num_episodes_per_update,), device = device) < self.action_dropout_prob
+
+            # roll out episode for a world model (optionally of a given gene) against environment
+
             for episode, gene_id in tqdm(self.episode_genes_for_process, desc = 'episodes' if not agent.evolutionary else 'episodes-genes', position = 1, disable = not is_main, leave = False):
 
                 latent_gene = None
@@ -1604,9 +1640,11 @@ class Learner(Module):
                     episode_seed = episode_seeds[episode]
                     reset_kwargs.update(seed = episode_seed.item())
 
-                one_episode_memories = deque([])
+                one_episode_memories = []
 
                 reward_dropout = reward_dropouts[episode]
+                prev_action_dropout = action_dropouts[episode]
+
                 reset_out = env.reset(**reset_kwargs)
 
                 if isinstance(reset_out, tuple):
@@ -1632,8 +1670,7 @@ class Learner(Module):
 
                     state_with_reward, inverse_pack = pack_with_inverse((state, prev_reward), '*')
 
-                    self.agent.rsnorm.eval()
-                    normed_state_reward = self.agent.rsnorm(state_with_reward)
+                    normed_state_reward = self.agent.rsnorm.forward_eval(state_with_reward)
 
                     normed_state, normed_reward = inverse_pack(normed_state_reward)
 
@@ -1649,6 +1686,7 @@ class Learner(Module):
                         normed_state,
                         rewards = normed_reward,
                         reward_dropout = reward_dropout,
+                        prev_action_dropout = prev_action_dropout,
                         latent_gene = latent_gene,
                         cache = world_model_cache,
                         actions = prev_action,
@@ -1744,6 +1782,7 @@ class Learner(Module):
                 episode_lens.append(timestep + 1)
                 gene_ids.append(gene_id)
                 one_update_reward_dropouts.append(reward_dropout)
+                one_update_action_dropouts.append(prev_action_dropout)
 
                 # add list[Memory] to all episode memories list[list[Memory]]
 
@@ -1761,6 +1800,7 @@ class Learner(Module):
                 tensor(episode_lens, device = device),
                 tensor(gene_ids, device = device),
                 stack(one_update_reward_dropouts),
+                stack(one_update_action_dropouts),
                 fitnesses
             )
 
@@ -1768,6 +1808,7 @@ class Learner(Module):
             episode_lens.clear()
             gene_ids.clear()
             one_update_reward_dropouts.clear()
+            one_update_action_dropouts.clear()
 
             if divisible_by(learning_update, self.save_every):
                 self.agent.save()
